#!/usr/bin/env python
# coding: utf-8

# Copyright (c) cdr4eelz.
# Distributed under the terms of the Modified BSD License.

"""Tests of custom Jupyter widget."""

#import pytest

#from serialhub import _jupyter_labextension_paths, SerialHubWidget
#from ..backend import SerialHubWidget
#from .. import XYZ

def test_backend_creation_blank():
    """Ensure traitlets exist and have reasonable defaults."""
    #pylint: disable=protected-access
    from .. import SerialHubWidget  # pylint: disable=import-outside-toplevel
    shw = SerialHubWidget()
    assert not shw.is_supported  # Ensure either None or False
    assert shw.status == 'Checking...'
    assert shw.value == ''
    assert isinstance(shw.request_options, dict)
    assert isinstance(shw.serial_options, dict)
    assert shw.pkt_send_front == shw.pkt_send_back == (0, 0)
    assert shw.pkt_recv_front == shw.pkt_recv_back == (0, 0)
    #Confirm some semi-internal settings under our control
    assert shw._model_name == 'SerialHubModel'
    assert shw._model_module == 'serialhub'
    assert isinstance(shw._model_module_version, str)
    assert shw._view_name == 'SerialHubView'
    assert shw._view_module == 'serialhub'
    assert isinstance(shw._view_module_version, str)

def test_backend_labextension():
    """Check that we seem to be a JupyterLab extension."""
    from .. import _jupyter_labextension_paths # pylint: disable=import-outside-toplevel
    # Ensure that it can be called without incident:
    path = _jupyter_labextension_paths()
    assert path
    # Some sanity checks:
    assert len(path) == 1
    assert isinstance(path[0], dict)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) cdr4eelz.
# Distributed under the terms of the Modified BSD License.

"""SerialHub backend widget & support classes"""

from __future__ import absolute_import
#from __future__ import print_function
from typing import Sequence, Mapping, Any, ByteString, Optional  # , BinaryIO, IO

import ipywidgets #from ipywidgets import DOMWidget, register
import traitlets
#from traitlets.traitlets import Integer #from traitlets import Unicode, Int, Bool, Complex, Enum
from ._frontend import module_name, module_version

@ipywidgets.register
class SerialHubWidget(ipywidgets.DOMWidget):
    """\
    SerialHubWidget class inherits ipywidgets.DOMWidget
        Model: SerialHubModel, View: SerialHubView
    """
    _model_name = traitlets.Unicode('SerialHubModel').tag(sync=True)
    _model_module = traitlets.Unicode(module_name).tag(sync=True)
    _model_module_version = traitlets.Unicode(module_version).tag(sync=True)
    _view_name = traitlets.Unicode('SerialHubView').tag(sync=True)
    _view_module = traitlets.Unicode(module_name).tag(sync=True)
    _view_module_version = traitlets.Unicode(module_version).tag(sync=True)

    #is_supported starts as None, then is set by the JavaScript front-end upon load
    is_supported = traitlets.Bool(
        allow_none=True,
        read_only=True,
        help='is_supported is set by frontend to True if the browser supports Web Serial API'
    ).tag(sync=True)
    #FUTURE: Consider an Enum or UseEnum approach for status traitlet???
    status = traitlets.Unicode(
        default_value='Checking...',
        help='status of the widget frontend regarding serial port'
    ).tag(sync=True)
    value = traitlets.Unicode(
        default_value='',
        help='value for debug feedback messages'
    ).tag(sync=True)
    request_options = traitlets.Dict(
        per_key_traits={
            'filters': traitlets.List(trait=traitlets.Dict())
        }, default_value={
            'filters': []
        }, help='request_options to filter serial port list by vendor:product codes'
    ).tag(sync=True)
    serial_options = traitlets.Dict(
        per_key_traits={
            'baudRate': traitlets.Int(),
            'dataBits': traitlets.Int(), #7 | 8
            'parity': traitlets.Unicode(), #'none' | 'even' | 'odd' (Use Enum???)
            'stopBits': traitlets.Int(), #1 | 2
            'bufferSize': traitlets.Int(), #default 255
            'flowControl': traitlets.Unicode() #'none' | 'hardware'
        }, default_value={
            'baudRate': 9600,
            'dataBits': 8,
            'parity': 'none',
            'stopBits': 1
        }, help='serial_options to apply when opening serial port'
    ).tag(sync=True)
    #FUTURE: Optionally turn off 'sync' attribute for statistics (for performance)
    pkt_recv_front = traitlets.Tuple(
        traitlets.Int(), traitlets.Int(),
        default_value=(0, 0),
        read_only=True,
        help='[bytes,packets] received by frontend JS reader'
    ).tag(sync=True)
    pkt_send_front = traitlets.Tuple(
        traitlets.Int(), traitlets.Int(),
        default_value=(0, 0),
        read_only=True,
        help='[bytes,packets] sent to serial port by frontend'
    ).tag(sync=True)
    pkt_recv_back = traitlets.Tuple(
        traitlets.Int(), traitlets.Int(),
        default_value=(0, 0),
        help='[bytes,packets] received from trontend by backend callback'
    ).tag(sync=True)
    pkt_send_back = traitlets.Tuple(
        traitlets.Int(), traitlets.Int(),
        default_value=(0, 0),
        help='[bytes,packets] sent by backend to frontend'
    ).tag(sync=True)

    def __init__(self, *args, **kwargs):
        #TODO: Allow request & serial options to be passed at construction
        ipywidgets.DOMWidget.__init__(self, *args, **kwargs)
        self._cb_recv = None
        self.on_msg(self.msg_custom)

    def on_recv(self, cb_recv):
        """Set callback for received buffers of serial data"""
        self._cb_recv = cb_recv

    def do_recv(self, buf: ByteString):
        """A buffer arrived, handle it."""
        if self._cb_recv:
            self._cb_recv(buf)
        else:
            # buf.hex() ; str(binascii.b2a_hex(buf)) ; buf.decode('ascii','ignore')
            decoded: str = str(buf, encoding='ascii', errors='ignore')
            self.value += decoded.replace("\n", "\\n").replace("\r", "\\r")

    def msg_custom(self,
                widget: ipywidgets.DOMWidget,  #pylint: disable=unused-argument
                content: Mapping[str, Any],
                buffers: Optional[Sequence[ByteString]] = None
    ) -> None:
        """Receives custom message callbacks from the frontend client."""
        msgtype = str(content['type'])
        if msgtype == 'RECV':
            for buf in buffers:
                (o_byt, o_pkt) = self.pkt_recv_back
                self.pkt_recv_back = (o_byt + len(buf), o_pkt + 1)
                self.do_recv(buf)
        elif msgtype == 'SENT': #Acknowledge data sent by client
            (f_byt, f_pkt) = content['stat_client'] #Update backend stats
            self.pkt_send_front = (f_byt, f_pkt)
        elif msgtype == 'RSTS': #Reset backend statistics
            self.pkt_recv_back = (0, 0)
            self.pkt_send_back = (0, 0)
        elif msgtype == 'MSGV': #Append to "value" from client
            self.value += content['text']

    def send_custom(self,
                content: Mapping[str, Any],
                buffers: Optional[Sequence[ByteString]] = None
    ) -> None:
        """Sends a custom message to frontend.
            Map "content" gets serialized to JSON available as JS object in client.
            Optional "buffers" get sent as binary data (unlike binary data within "content").
        """
        self.send(content, buffers)

    def write_bytes(self,
                data: ByteString
    ) -> None:
        """Sends a buffer for the client to write() as serial data.
          "data" is a single bytestring to be output by client to the serialport
        """
        self.send_custom({'type': 'SEND'}, [data])
        (o_byt, o_pkt) = self.pkt_send_back
        self.pkt_send_back = (o_byt + len(data), o_pkt + 1)

    def write_str(self,
                data: str,
                enc: Optional[str] = 'ascii',
                errs: Optional[str] = 'ignore'
    ) -> None:
        """Sends a single buffer for the client to write() as serial data.
            Optional "enc" for desired encoding (default 'ascii')
            Optional "errs" for desired encoding error handling (default 'ignore')
        """
        self.write_bytes(data.encode(encoding=enc, errors=errs))



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) cdr4eelz.
# Distributed under the terms of the Modified BSD License.

"""Wrapper class to make serialhub.SerialHubPort access similar to regular IO."""

# , BinaryIO, IO
from typing import Deque, ByteString, Optional, NoReturn
import io  # , binascii
from collections import deque

from .backend import SerialHubWidget

#BinaryIO(IO[bytes])
class SerialIO(io.RawIOBase):
    """Serial IO proxied to frontend browser serial"""

    def __init__(self, widget: SerialHubWidget):  # , *args, **kwargs):
        io.RawIOBase.__init__(self)
        self.widget: SerialHubWidget = widget
        self.bufseq: Deque[io.BytesIO] = deque()

        def wrap_recv(buf):
            self.cb_recv(buf)  # Capture self in a callback wrapper
        widget.on_recv(wrap_recv)

    def cb_recv(self, buf: ByteString) -> None:
        """Append received data to our deque as BytesIO."""
        bio = io.BytesIO(buf)
        self.bufseq.append(bio)

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def isatty(self) -> bool:
        return False

    def seekable(self) -> bool:
        return False

    def closed(self) -> bool:
        """Currently True only if widget reports unsupported"""
        return not self.widget.is_supported or self.widget.status != 'Connected'

    def write(self, data: bytes) -> Optional[int]:
        if self.closed():
            raise ValueError("Stream closed")
        if len(data) <= 0:
            return None
        self.widget.write_bytes(data)
        return len(data)

    def readinto(self, b: bytearray) -> Optional[int]:
        if self.closed():
            raise ValueError("Stream closed")
        if len(b) <= 0:
            return None
        if len(self.bufseq) <= 0:
            return 0
        return self.bufseq[0].readinto(b)

    def fileno(self) -> NoReturn:
        raise OSError("No fileno")

    def tell(self) -> NoReturn:
        raise OSError("Not seekable")

    def seek(self, offset: int, whence=0) -> NoReturn:
        raise OSError("Not seekable")

    def truncate(self, size: Optional[int] = None) -> NoReturn:
        raise OSError("Not seekable")



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) cdr4eelz.
# Distributed under the terms of the Modified BSD License.

"""
Information about the frontend package of the widgets.
"""

module_name = "serialhub" #pylint: disable=invalid-name
module_version = "^0.0.35"  # pylint: disable=invalid-name



# Copyright (c) cdr4eelz.
# Distributed under the terms of the Modified BSD License.

"""\
Python backend module for SerialHub Widget
"""

import json
from pathlib import Path

#Import these so they get re-exported to serialhub package
from .backend import SerialHubWidget
from .serialio import SerialIO
from .nbextension import _jupyter_nbextension_paths

from ._version import __version__

HERE = Path(__file__).parent.resolve()

with (HERE / "labextension" / "package.json").open() as fid:
    data = json.load(fid)

def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": data["name"]
    }]
